/*
 * header.c
 *
 * Read, write, or create a WIM header.
 */

/*
 * Copyright (C) 2012, 2013 Eric Biggers
 *
 * This file is part of wimlib, a library for working with WIM files.
 *
 * wimlib is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option)
 * any later version.
 *
 * wimlib is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wimlib; if not, see http://www.gnu.org/licenses/.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "wimlib.h"
#include "wimlib/assert.h"
#include "wimlib/endianness.h"
#include "wimlib/error.h"
#include "wimlib/file_io.h"
#include "wimlib/header.h"
#include "wimlib/util.h"
#include "wimlib/wim.h"

#include <limits.h>
#include <string.h>
#include <unistd.h>
#ifdef HAVE_ALLOCA_H
#  include <alloca.h>
#else
#  include <stdlib.h>
#endif

/*
 * Reads the header from a WIM file.
 *
 * @filename
 *	Name of the WIM file (for error/debug messages only), or NULL if reading
 *	directly from file descriptor.
 * @in_fd
 *	File descriptor, positioned at offset 0, to read the header from.
 * @hdr
 *	Structure to read the header into.
 *
 * Return values:
 *	WIMLIB_ERR_SUCCESS (0)
 *	WIMLIB_ERR_IMAGE_COUNT
 *	WIMLIB_ERR_INVALID_CHUNK_SIZE
 *	WIMLIB_ERR_INVALID_PART_NUMBER
 *	WIMLIB_ERR_NOT_A_WIM_FILE
 *	WIMLIB_ERR_READ
 *	WIMLIB_ERR_UNEXPECTED_END_OF_FILE
 *	WIMLIB_ERR_UNKNOWN_VERSION
 */
int
read_wim_header(const tchar *filename, struct filedes *in_fd,
		struct wim_header *hdr)
{
	struct wim_header_disk disk_hdr _aligned_attribute(8);
	int ret;
	tchar *pipe_str;

	wimlib_assert(in_fd->offset == 0);

	if (!filename) {
		pipe_str = alloca(40);
		tsprintf(pipe_str, T("[fd %d]"), in_fd->fd);
		filename = pipe_str;
	}

	BUILD_BUG_ON(sizeof(struct wim_header_disk) != WIM_HEADER_DISK_SIZE);

	DEBUG("Reading WIM header from \"%"TS"\"", filename);

	ret = full_read(in_fd, &disk_hdr, sizeof(disk_hdr));
	if (ret)
		goto read_error;

	if (disk_hdr.magic != WIM_MAGIC) {
		if (disk_hdr.magic == PWM_MAGIC) {
			/* Pipable WIM:  Use header at end instead, unless
			 * actually reading from a pipe.  */
			if (!in_fd->is_pipe) {
				lseek(in_fd->fd, -WIM_HEADER_DISK_SIZE, SEEK_END);
				ret = full_read(in_fd, &disk_hdr, sizeof(disk_hdr));
				if (ret)
					goto read_error;
			}
		} else {
			ERROR("\"%"TS"\": Invalid magic characters in header", filename);
			return WIMLIB_ERR_NOT_A_WIM_FILE;
		}
	}
	hdr->magic = disk_hdr.magic;

	if (le32_to_cpu(disk_hdr.hdr_size) != sizeof(struct wim_header_disk)) {
		ERROR("\"%"TS"\": Header size is invalid (%u bytes)",
		      filename, le32_to_cpu(disk_hdr.hdr_size));
		return WIMLIB_ERR_INVALID_HEADER;
	}

	if (le32_to_cpu(disk_hdr.wim_version) != WIM_VERSION) {
		ERROR("\"%"TS"\": The WIM header says the WIM version is %u, "
		      "but wimlib only knows about version %u",
		      filename, le32_to_cpu(disk_hdr.wim_version), WIM_VERSION);
		if (le32_to_cpu(disk_hdr.wim_flags) & WIM_HDR_FLAG_COMPRESS_LZMS) {
			ERROR("\"%"TS"\": This WIM uses LZMS compression, "
			      "which is not supported by wimlib.", filename);
		}
		return WIMLIB_ERR_UNKNOWN_VERSION;
	}

	hdr->flags = le32_to_cpu(disk_hdr.wim_flags);
	if (le32_to_cpu(disk_hdr.chunk_size) != WIM_CHUNK_SIZE &&
	    (hdr->flags & WIM_HDR_FLAG_COMPRESSION)) {
		ERROR("\"%"TS"\": Unexpected chunk size of %u! Ask the author to "
		      "implement support for other chunk sizes.",
		      filename, le32_to_cpu(disk_hdr.chunk_size));
		ERROR("(Or it might just be that the WIM header is invalid.)");
		return WIMLIB_ERR_INVALID_CHUNK_SIZE;
	}

	memcpy(hdr->guid, disk_hdr.guid, WIM_GID_LEN);

	hdr->part_number = le16_to_cpu(disk_hdr.part_number);
	hdr->total_parts = le16_to_cpu(disk_hdr.total_parts);

	if (hdr->total_parts == 0 || hdr->part_number == 0 ||
	    hdr->part_number > hdr->total_parts)
	{
		ERROR("\"%"TS"\": Invalid WIM part number: %hu of %hu",
		      filename, hdr->part_number, hdr->total_parts);
		return WIMLIB_ERR_INVALID_PART_NUMBER;
	}

	hdr->image_count = le32_to_cpu(disk_hdr.image_count);

	DEBUG("part_number = %u, total_parts = %u, image_count = %u",
	      hdr->part_number, hdr->total_parts, hdr->image_count);

	if (hdr->image_count >= INT_MAX) {
		ERROR("\"%"TS"\": Invalid image count (%u)",
		      filename, hdr->image_count);
		return WIMLIB_ERR_IMAGE_COUNT;
	}

	get_resource_entry(&disk_hdr.lookup_table_res_entry, &hdr->lookup_table_res_entry);
	get_resource_entry(&disk_hdr.xml_data_res_entry, &hdr->xml_res_entry);
	get_resource_entry(&disk_hdr.boot_metadata_res_entry, &hdr->boot_metadata_res_entry);
	hdr->boot_idx = le32_to_cpu(disk_hdr.boot_idx);
	get_resource_entry(&disk_hdr.integrity_table_res_entry, &hdr->integrity);
	return 0;

read_error:
	ERROR_WITH_ERRNO("\"%"TS"\": Error reading header", filename);
	return ret;
}

/* Writes the header for a WIM file at the specified offset.  If the offset
 * specified is the current one, the position is advanced by the size of the
 * header.  */
int
write_wim_header_at_offset(const struct wim_header *hdr, struct filedes *out_fd,
			   off_t offset)
{
	struct wim_header_disk disk_hdr _aligned_attribute(8);
	int ret;

	DEBUG("Writing %sWIM header at offset %"PRIu64,
	      ((hdr->magic == PWM_MAGIC) ? "pipable " : ""),
	      offset);

	disk_hdr.magic = hdr->magic;
	disk_hdr.hdr_size = cpu_to_le32(sizeof(struct wim_header_disk));
	disk_hdr.wim_version = cpu_to_le32(WIM_VERSION);
	disk_hdr.wim_flags = cpu_to_le32(hdr->flags);
	disk_hdr.chunk_size = cpu_to_le32((hdr->flags & WIM_HDR_FLAG_COMPRESSION) ?
					  	WIM_CHUNK_SIZE : 0);
	memcpy(disk_hdr.guid, hdr->guid, WIM_GID_LEN);

	disk_hdr.part_number = cpu_to_le16(hdr->part_number);
	disk_hdr.total_parts = cpu_to_le16(hdr->total_parts);
	disk_hdr.image_count = cpu_to_le32(hdr->image_count);
	put_resource_entry(&hdr->lookup_table_res_entry, &disk_hdr.lookup_table_res_entry);
	put_resource_entry(&hdr->xml_res_entry, &disk_hdr.xml_data_res_entry);
	put_resource_entry(&hdr->boot_metadata_res_entry, &disk_hdr.boot_metadata_res_entry);
	disk_hdr.boot_idx = cpu_to_le32(hdr->boot_idx);
	put_resource_entry(&hdr->integrity, &disk_hdr.integrity_table_res_entry);
	memset(disk_hdr.unused, 0, sizeof(disk_hdr.unused));

	if (offset == out_fd->offset)
		ret = full_write(out_fd, &disk_hdr, sizeof(disk_hdr));
	else
		ret = full_pwrite(out_fd, &disk_hdr, sizeof(disk_hdr), offset);
	if (ret)
		ERROR_WITH_ERRNO("Failed to write WIM header");
	return ret;
}

/* Writes the header for a WIM file at the output file descriptor's current
 * offset.  */
int
write_wim_header(const struct wim_header *hdr, struct filedes *out_fd)
{
	return write_wim_header_at_offset(hdr, out_fd, out_fd->offset);
}

/* Update just the wim_flags field. */
int
write_wim_header_flags(u32 hdr_flags, struct filedes *out_fd)
{
	le32 flags = cpu_to_le32(hdr_flags);

	return full_pwrite(out_fd, &flags, sizeof(flags),
			   offsetof(struct wim_header_disk, wim_flags));
}

u32
get_wim_hdr_cflags(int ctype)
{
	switch (ctype) {
	case WIMLIB_COMPRESSION_TYPE_NONE:
		return 0;
	case WIMLIB_COMPRESSION_TYPE_LZX:
		return WIM_HDR_FLAG_COMPRESSION | WIM_HDR_FLAG_COMPRESS_LZX;
	case WIMLIB_COMPRESSION_TYPE_XPRESS:
		return WIM_HDR_FLAG_COMPRESSION | WIM_HDR_FLAG_COMPRESS_XPRESS;
	case WIMLIB_COMPRESSION_TYPE_INVALID:
		break;
	}
	return (u32)~0U;
}

/*
 * Initializes the header for a WIM file.
 */
int
init_wim_header(struct wim_header *hdr, int ctype)
{
	memset(hdr, 0, sizeof(struct wim_header));
	hdr->flags = get_wim_hdr_cflags(ctype);
	if (hdr->flags == (u32)~0U) {
		ERROR("Invalid compression type specified (%d)", ctype);
		return WIMLIB_ERR_INVALID_COMPRESSION_TYPE;
	}
	hdr->total_parts = 1;
	hdr->part_number = 1;
	randomize_byte_array(hdr->guid, sizeof(hdr->guid));
	hdr->magic = WIM_MAGIC;
	return 0;
}

struct hdr_flag {
	u32 flag;
	const char *name;
};
struct hdr_flag hdr_flags[] = {
	{WIM_HDR_FLAG_RESERVED, 	"RESERVED"},
	{WIM_HDR_FLAG_COMPRESSION,	"COMPRESSION"},
	{WIM_HDR_FLAG_READONLY,		"READONLY"},
	{WIM_HDR_FLAG_SPANNED,		"SPANNED"},
	{WIM_HDR_FLAG_RESOURCE_ONLY,	"RESOURCE_ONLY"},
	{WIM_HDR_FLAG_METADATA_ONLY,	"METADATA_ONLY"},
	{WIM_HDR_FLAG_WRITE_IN_PROGRESS,"WRITE_IN_PROGRESS"},
	{WIM_HDR_FLAG_RP_FIX,		"RP_FIX"},
	{WIM_HDR_FLAG_COMPRESS_RESERVED,"COMPRESS_RESERVED"},
	{WIM_HDR_FLAG_COMPRESS_LZX,	"COMPRESS_LZX"},
	{WIM_HDR_FLAG_COMPRESS_XPRESS,	"COMPRESS_XPRESS"},
};

/* API function documented in wimlib.h  */
WIMLIBAPI void
wimlib_print_header(const WIMStruct *wim)
{
	const struct wim_header *hdr = &wim->hdr;

	tprintf(T("Magic Characters            = MSWIM\\000\\000\\000\n"));
	tprintf(T("Header Size                 = %u\n"), WIM_HEADER_DISK_SIZE);
	tprintf(T("Version                     = 0x%x\n"), WIM_VERSION);

	tprintf(T("Flags                       = 0x%x\n"), hdr->flags);
	for (size_t i = 0; i < ARRAY_LEN(hdr_flags); i++)
		if (hdr_flags[i].flag & hdr->flags)
			tprintf(T("    WIM_HDR_FLAG_%s is set\n"), hdr_flags[i].name);

	tprintf(T("Chunk Size                  = %u\n"), WIM_CHUNK_SIZE);
	tfputs (T("GUID                        = "), stdout);
	print_byte_field(hdr->guid, WIM_GID_LEN, stdout);
	tputchar(T('\n'));
	tprintf(T("Part Number                 = %hu\n"), wim->hdr.part_number);
	tprintf(T("Total Parts                 = %hu\n"), wim->hdr.total_parts);
	tprintf(T("Image Count                 = %u\n"), hdr->image_count);
	tprintf(T("Lookup Table Size           = %"PRIu64"\n"),
				(u64)hdr->lookup_table_res_entry.size);
	tprintf(T("Lookup Table Flags          = 0x%hhx\n"),
				(u8)hdr->lookup_table_res_entry.flags);
	tprintf(T("Lookup Table Offset         = %"PRIu64"\n"),
				hdr->lookup_table_res_entry.offset);
	tprintf(T("Lookup Table Original_size  = %"PRIu64"\n"),
				hdr->lookup_table_res_entry.original_size);
	tprintf(T("XML Data Size               = %"PRIu64"\n"),
				(u64)hdr->xml_res_entry.size);
	tprintf(T("XML Data Flags              = 0x%hhx\n"),
				(u8)hdr->xml_res_entry.flags);
	tprintf(T("XML Data Offset             = %"PRIu64"\n"),
				hdr->xml_res_entry.offset);
	tprintf(T("XML Data Original Size      = %"PRIu64"\n"),
				hdr->xml_res_entry.original_size);
	tprintf(T("Boot Metadata Size          = %"PRIu64"\n"),
				(u64)hdr->boot_metadata_res_entry.size);
	tprintf(T("Boot Metadata Flags         = 0x%hhx\n"),
				(u8)hdr->boot_metadata_res_entry.flags);
	tprintf(T("Boot Metadata Offset        = %"PRIu64"\n"),
				hdr->boot_metadata_res_entry.offset);
	tprintf(T("Boot Metadata Original Size = %"PRIu64"\n"),
				hdr->boot_metadata_res_entry.original_size);
	tprintf(T("Boot Index                  = %u\n"), hdr->boot_idx);
	tprintf(T("Integrity Size              = %"PRIu64"\n"),
				(u64)hdr->integrity.size);
	tprintf(T("Integrity Flags             = 0x%hhx\n"),
				(u8)hdr->integrity.flags);
	tprintf(T("Integrity Offset            = %"PRIu64"\n"),
				hdr->integrity.offset);
	tprintf(T("Integrity Original_size     = %"PRIu64"\n"),
				hdr->integrity.original_size);
}
